package parsec3

import assertThrows
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.flow.receiveAsFlow
import kotlinx.coroutines.launch
import kotlinx.coroutines.test.runTest
import net.sergeych.parsec3.*
import kotlin.test.Test
import kotlin.test.assertEquals
import kotlin.test.assertTrue

internal class AdapterTest {

    object Api1 : CommandHost<WithAdapter>() {
        // create command `foo` that takes a string argument and
        // returns a string:
        val foo by command<String, String>()
    }


    object Api2 : CommandHost<WithAdapter>() {

        val bar by command<String, String>()
    }

    @Test
    fun interconnect() = runTest {
        val ch12 = Channel<ByteArray>()
        val ch21 = Channel<ByteArray>()
        val api1 = Api1
        val api2 = Api2
        api1.on(api1.foo) {
            it + "foo"
        }
        api2.on(api2.bar) {
            it + "bar"
        }

        val a1 = Adapter(WithAdapter(), api1) { ch12.send(it) }
        val a2 = Adapter(WithAdapter(), api2) { ch21.send(it) }
        launch { for (b in ch12) a2.receiveFrame(b) }
        launch { for (b in ch21) a1.receiveFrame(b) }

        assertEquals("123bar", a1.invokeCommand(api2.bar, "123"))
        assertEquals("321foo", a2.invokeCommand(api1.foo, "321"))

        ch12.cancel()
        ch21.cancel()

    }

    data class TestSession(var buzz: String) : WithAdapter()

    object ApiS1 : CommandHost<TestSession>() {
        // create command `foo` that takes a string argument and
        // returns a string:
        val foo by command<String, String>()
        val ex by command<String,Unit>()
        val ex2 by command<String,Unit>()
        val nullableTest by command<Int,String?>()
    }
    class ApiS2<T: WithAdapter> : CommandHost<T>() {
        // create command `foo` that takes a string argument and
        // returns a string:
        val bar by command<String, String>()
        val loopCall by command<String, String>()
    }

    @Test
    fun builderTest() = runTest {
        val ch12 = Channel<ByteArray>()
        val ch21 = Channel<ByteArray>()

        val er = ExceptionsRegistry().also {
            it.register("foo_x") { IllegalArgumentException(it) }
        }

        var b1Cancelled = false
        var b2Cancelled = false
        val b1 = AdapterBuilder(ApiS1, er) {
            newSession { TestSession("42") }
            on(api.foo) {
                it + buzz + "foo"
            }
            on(ApiS1.ex) {
                throw ParsecException("foo_x", it)
            }
            on(ApiS1.ex2) {
                throw IllegalArgumentException()
            }
            on(ApiS1.nullableTest) {
                if( it == 1 ) "one" else null
            }
            onCancel {
                b1Cancelled = true
            }
        }
        val b2 = AdapterBuilder(ApiS2(), er) {
            on(api.bar) {
                it + "bar"
            }
            on(api.loopCall) {
                try {
                    val res = ApiS1.foo(adapter, "loop-")
                    "%% $res %%"
                }
                catch(t: Throwable) {
                    t.printStackTrace()
                    throw t
                }
            }
            onCancel { b2Cancelled = true }
        }

        val a1 = b1.createWith(ch21.receiveAsFlow()) {
            ch12.send(it)
        }
        val a2 = b2.createWith(ch12.receiveAsFlow()) {
            ch21.send(it)
        }

//        assertEquals("123bar", a1.invokeCommand(ApiS2<Unit>().bar, "123"))
        assertEquals("%% loop-42foo %%", a1.invokeCommand(ApiS2<WithAdapter>().loopCall, "123"))
        assertEquals("32142foo", a2.invokeCommand(ApiS1.foo, "321"))

        assertEquals("one", a2.invokeCommand(ApiS1.nullableTest,1))
        assertEquals(null, a2.invokeCommand(ApiS1.nullableTest,2))

        assertEquals("---42foo", ApiS1.foo.invoke(a2, "---"))
        val x = assertThrows<IllegalArgumentException> { ApiS1.ex.invoke(a2, "foobar") }
        assertEquals("foobar", x.message)
        assertThrows<IllegalArgumentException> { ApiS1.ex2.invoke(a2, "foobar") }

        ch12.cancel()
        ch21.cancel()
        a1.cancel()
        a2.cancel()

        assertTrue { b1Cancelled }
        assertTrue { b2Cancelled }
    }
    @Test
    fun builderDupeErrorTest() = runTest {
        assertThrows<DuplicateCommandDefinition> {
            AdapterBuilder(ApiS1) {
                newSession { TestSession("42") }
                on(api.foo) {
                    it + buzz + "foo"
                }
                on(api.foo) {
                    it + buzz + "foo"
                }
            }
        }
    }
}